from hivemind_bus_client.message import HiveMessage, HiveMessageType
from hivemind_core.protocol import HiveMindListenerProtocol, HiveMindClientConnection, \
    HiveMindListenerInternalProtocol
from ovos_bus_client import Message
from ovos_core.intent_services import IntentService
from ovos_utils.log import LOG

from local_hive.fakebus import FakeBus


class LocalHiveInternalProtocol(HiveMindListenerInternalProtocol):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.intent_service = None
        self.intent2skill = {}
        self.permission_overrides = {}

    def register_bus_handlers(self):
        LOG.info("registering intent service bus handlers")
        self.intent_service = IntentService(self.bus)
        self.bus.on("message", self.handle_internal_mycroft)  # catch all

    def skill2peer(self, skill_id):
        for peer, client in self.clients.items():
            if client.key == skill_id:
                return peer
        return None

    # mycroft handlers  - from LocalHive -> skill
    def handle_internal_mycroft(self, message: str):
        """ forward internal messages to clients if they are the target
        here is where the client isolation happens,
        clients only get responses to their own messages"""

        # "message" event is a special case in ovos-bus-client that is not deserialized
        message = Message.deserialize(message)

        skill_id = message.context.get("skill_id")
        peers = message.context.get("destination") or []

        # converse method handling
        if message.msg_type in ["skill.converse.request"]:
            skill_id = message.data.get("skill_id")
            message.context["skill_id"] = skill_id
            skill_peer = self.skill2peer(skill_id)
            LOG.info(f"Converse: {message.msg_type} "
                     f"Skill: {skill_id} "
                     f"Peer: {skill_peer}")
            message.context['source'] = "IntentService"
            message.context['destination'] = peers
            client = self.clients[skill_peer]
            client.send(
                HiveMessage(HiveMessageType.BUS, message)
            )

        elif message.msg_type in ["skill.converse.response"]:
            # just logging that it was received, converse method handled by
            # skill
            skill_id = message.data.get("skill_id")
            response = message.data.get("result")
            message.context["skill_id"] = skill_id
            skill_peer = self.skill2peer(skill_id)
            LOG.info(f"Converse Response: {response} "
                     f"Skill: {skill_id} "
                     f"Peer: {skill_peer}")
            message.context['source'] = skill_id
            message.context['destination'] = peers

        # intent found
        elif message.msg_type in self.intent2skill:
            skill_id = self.intent2skill[message.msg_type]
            skill_peer = self.skill2peer(skill_id)
            message.context["skill_id"] = skill_id

            LOG.info(f"Intent: {message.msg_type} "
                     f"Skill: {skill_id} "
                     f"Source: {peers} "
                     f"Target: {skill_peer}")

            # trigger the skill
            message.context['source'] = "IntentService"
            LOG.debug(f"Triggering intent: {skill_peer}")
            client = self.clients[skill_peer]
            client.send(
                HiveMessage(HiveMessageType.BUS, message)
            )

        # skill registering intent - keep track internally
        elif message.msg_type in ["register_intent",
                                  "padatious:register_intent"]:
            LOG.info(f"Register Intent: {message.data['name']} "
                     f"Skill: {message.context['skill_id']}")
            self.intent2skill[message.data["name"]] = skill_id


class LocalHiveProtocol(HiveMindListenerProtocol):
    intent_messages = [
        "recognizer_loop:utterance",
        "intent.service.intent.get",
        "intent.service.skills.get",
        "intent.service.active_skills.get",
        "intent.service.adapt.get",
        "intent.service.padatious.get",
        "intent.service.adapt.manifest.get",
        "intent.service.padatious.manifest.get",
        "intent.service.adapt.vocab.manifest.get",
        "intent.service.padatious.entities.manifest.get",
        "register_vocab",
        "register_intent",
        "detach_intent",
        "detach_skill",
        "add_context",
        "remove_context",
        "clear_context",
        "mycroft.skills.loaded",
        "active_skill_request",
        'mycroft.speech.recognition.unknown',
        'padatious:register_intent',
        'padatious:register_entity',
        'mycroft.skills.initialized'
    ]
    default_permissions = intent_messages + [
        "speak",
        "mycroft.skill.handler.start",
        "mycroft.skill.handler.complete",
        "skill.converse.request",
        "skill.converse.response"
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.bus = None
        self.intent2skill = {}
        self.permission_overrides = {}

    def handle_new_client(self, client: HiveMindClientConnection):
        LOG.info(f"new client: {client.peer}")
        self.clients[client.peer] = client
        message = Message("hive.client.connect",
                          {"ip": client.ip, "session_id": client.sess.session_id},
                          {"source": client.peer})
        self.internal_protocol.bus.emit(message)

        msg = HiveMessage(HiveMessageType.HELLO,
                          payload={"handshake": False,
                                   "crypto": False,
                                   "peer": client.peer,  # this identifies the connected client in ovos message.context
                                   "node_id": self.peer})
        LOG.info(f"saying HELLO to: {client.peer}")
        client.send(msg)

        # request client to start handshake (by sending client pubkey)
        payload = {
            "handshake": False,  # tell the client it must do a handshake or connection will be dropped
            "binarize": False,  # report we support the binarization scheme
            "preshared_key": False,  # do we have a pre-shared key (V0 proto)
            "password": False,  # is password available (V1 proto, replaces pre-shared key)
            "crypto_required": False  # do we allow unencrypted payloads
        }
        msg = HiveMessage(HiveMessageType.HANDSHAKE, payload)
        LOG.info(f"starting {client.peer} HANDSHAKE: {payload}")
        client.send(msg)
        # if client is in protocol V1 -> self.handle_handshake_message
        # clients can rotate their pubkey or session_key by sending a new handshake

    def handle_handshake_message(self, message: HiveMessage,
                                 client: HiveMindClientConnection):
        LOG.info("handshake received, ignoring")

    @property
    def intent_service(self):
        return self.internal_protocol.intent_service

    def bind(self, websocket, bus=None):
        websocket.protocol = self
        if bus is None:
            bus = FakeBus()
        self.bus = bus
        self.internal_protocol = LocalHiveInternalProtocol(self.bus)
        self.internal_protocol.register_bus_handlers()

    # messages from skill -> LocalHive
    def handle_inject_mycroft_msg(self, message: Message, client: HiveMindClientConnection):
        """
        message (Message): mycroft bus message object
        """
        # message from a terminal
        if message.msg_type == "recognizer_loop:utterance":
            LOG.info(f"Utterance: {message.data['utterances']} "
                     f"Peer: {client.peer}")
            message.context["source"] = client.peer
            self.intent_service.handle_utterance(message)

        # message from a skill
        elif message.context.get("skill_id"):
            message.context["source"] = client.peer
            self.handle_skill_message(message)

    def handle_skill_message(self, message):
        """ message sent by skill"""
        if isinstance(message, str):
            message = Message.deserialize(message)

        skill_id = message.context.get("skill_id")
        intent_skill = self.intent2skill.get(message.msg_type)
        permitted = False

        # skill intents
        if intent_skill:
            permitted = True
        # skill_id permission override
        elif skill_id and skill_id in self.permission_overrides:
            if message.msg_type in self.permission_overrides[skill_id]:
                permitted = True
        # default permissions
        elif message.msg_type in self.default_permissions:
            permitted = True

        if permitted:
            peers = message.context.get('destination') or []
            if isinstance(peers, str):
                peers = [peers]

            # check if it should be forwarded to some peer (skill/terminal)
            for peer in peers:
                client = self.clients.get(peer)
                if client:
                    LOG.debug(f"destination: {message.context['destination']} "
                              f"skill:{skill_id} "
                              f"type:{message.msg_type}")
                    client.send(
                        HiveMessage(HiveMessageType.BUS, message)
                    )

            # check if this message should be forwarded to intent service
            if message.msg_type in self.intent_messages or "IntentService" in peers:
                self.bus.emit(message)
        else:
            self.handle_ignored_message(message)

    def handle_ignored_message(self, message):
        pass
